def pytest_markdown_docs_markdown_it():
    """Support admonitions."""
    import markdown_it
    from mdit_py_plugins.admon import admon_plugin

    mi = markdown_it.MarkdownIt(config="commonmark")
    mi.use(admon_plugin)
    return mi



"""Plotting utilities for visualizing GAM models."""

import types
from collections.abc import Callable, Mapping
from copy import deepcopy
from dataclasses import dataclass
from math import ceil
from typing import Any, Literal, TypeGuard

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib import rcParams
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from pandas import CategoricalDtype
from pandas.api.types import is_numeric_dtype

from pymgcv.basis_functions import FactorSmooth, RandomEffect
from pymgcv.gam import AbstractGAM
from pymgcv.qq import QQResult, qq_simulate
from pymgcv.rlibs import rstats
from pymgcv.rpy_utils import to_py
from pymgcv.terms import (
    AbstractTerm,
    L,
    S,
    T,
    _FactorSmoothToByInterface,
)
from pymgcv.utils import data_len


def plot_gam(
    gam: AbstractGAM,
    *,
    ncols: int = 2,
    scatter: bool = False,
    data: pd.DataFrame | Mapping[str, np.ndarray | pd.Series] | None = None,
    to_plot: type | types.UnionType | dict[str, list[AbstractTerm]] = AbstractTerm,
    kwargs_mapper: dict[Callable, dict[str, Any]] | None = None,
) -> tuple[Figure, Axes | np.ndarray]:
    """Plot a gam model.

    Except for some specialised cases, this plots the partial effects of the terms.

    Args:
        gam: The fitted gam object to plot.
        ncols: The number of columns before wrapping axes.
        scatter: Whether to plot the residuals (where possible), and the overlayed
            datapoints on 2D plots. For more fine control, see `kwargs_mapper`. Defaults
            to False.
        data: The data to use for plotting partial residuals and scatter points.
            Will default to using the data used for fitting. Only relevant if
            `scatter=True`.
        to_plot: Which terms to plot. If a type, only plots terms
            of that type (e.g. ``to_plot = S | T`` to plot smooths).
            If a dictionary, it should map the target names to
            an iterable of terms to plot (similar to how models are specified).
        kwargs_mapper: Used to pass keyword arguments to the underlying `pymgcv.plot`
            functions. A dictionary mapping the plotting function to kwargs. For
            example, to disable the confidence intervals on the 1d plots, set
            ``kwargs_mapper`` to
            ```python
            from pymgcv.plot import plot_continuous_1d
            {plot_continuous_1d: {"fill_between_kwargs": {"disable": True}}}
            ```
    """
    if gam.fit_state is None:
        raise ValueError("Cannot plot before fitting the model.")

    kwargs_mapper = {} if kwargs_mapper is None else kwargs_mapper
    kwargs_mapper.setdefault(plot_categorical, {}).setdefault("residuals", scatter)
    kwargs_mapper.setdefault(plot_continuous_1d, {}).setdefault("residuals", scatter)
    kwargs_mapper.setdefault(plot_continuous_2d, {}).setdefault(
        "scatter_kwargs",
        {},
    ).setdefault("disable", not scatter)

    if isinstance(to_plot, type | types.UnionType):
        to_plot = {
            k: [v for v in terms if isinstance(v, to_plot)]
            for k, terms in gam.all_predictors.items()
        }

    plotters = []
    for target, terms in to_plot.items():
        for term in terms:
            try:
                plotter = get_term_plotter(
                    term=term,
                    gam=gam,
                    target=target,
                    data=data,
                )
            except NotImplementedError:
                continue
            plotters.append(plotter)

    n_axs = sum(p.required_axes for p in plotters)
    if n_axs == 0:
        raise ValueError("Do not know how to plot any terms in the model.")

    ncols = min(n_axs, ncols)
    fig, axes = plt.subplots(
        nrows=ceil(n_axs / ncols),
        ncols=ncols,
        layout="constrained",
    )
    if not isinstance(axes, np.ndarray):
        axes = np.array([axes])
    axes = axes.ravel()

    idx = 0
    for plotter in plotters:
        kwargs = kwargs_mapper.get(plotter.underlying_function, {})
        plotter.make_plot(axes[idx : (idx + plotter.required_axes)], **kwargs)

        # If multiple targets in the model, add target variable as title.
        if len(gam.all_predictors) > 1:
            for ax in axes[idx : (idx + plotter.required_axes)]:
                ax.set_title(plotter.target_variable)

        idx += plotter.required_axes

    # Hide unused axes
    for ax in axes[idx:]:
        ax.set_axis_off()

    return fig, axes


@dataclass
class _TermPlotter:
    make_plot: Callable
    underlying_function: Callable
    target_variable: str
    required_axes: int = 1


def get_term_plotter(
    term: AbstractTerm,
    gam: AbstractGAM,
    target: str,
    data: pd.DataFrame | Mapping[str, np.ndarray | pd.Series] | None = None,
) -> _TermPlotter:
    """Utility for plotting a term in a model.

    Because some terms need multiple axes for plotting, this returns the number of axes
    required, and a function that applies the plotting to an iterable of axes, taking
    the axes and **kwargs passed to the plotting function. This allows us to setup the
    axes before plotting when plotting multiple terms.
    """
    if gam.fit_state is None:
        raise ValueError("Cannot plot before fitting the model.")
    data = data if data is not None else gam.fit_state.data
    data = deepcopy(data)

    if _is_random_wiggly(term):
        term = _FactorSmoothToByInterface(term)

    dtypes = {k: data[k].dtype for k in term.varnames}
    by_dtype = data[term.by].dtype if term.by is not None else None
    if isinstance(by_dtype, CategoricalDtype):
        levels = by_dtype.categories
    else:
        levels = [None]
    dim = len(term.varnames)

    def _all_numeric(dtypes: dict):
        return all(is_numeric_dtype(dtype) for dtype in dtypes.values())

    match (dim, term):
        case (1, L()) if isinstance(dtypes[term.varnames[0]], CategoricalDtype):

            def _plot_wrapper(axes, **kwargs):
                axes[0] = plot_categorical(
                    term=term,
                    gam=gam,
                    target=target,
                    data=data,
                    ax=axes[0],
                    **kwargs,
                )
                return axes

            return _TermPlotter(_plot_wrapper, plot_categorical, target)

        case (1, S()) if isinstance(term.bs, RandomEffect):

            def _plot_wrapper(axes, **kwargs):
                axes[0] = plot_random_effect(
                    term=term,
                    gam=gam,
                    target=target,
                    ax=axes[0],
                    **kwargs,
                )
                return axes

            return _TermPlotter(_plot_wrapper, plot_random_effect, target)

        # TODO "re" basis?

        case (1, AbstractTerm()) if _all_numeric(dtypes):

            def _plot_wrapper(axes, **kwargs):
                for level in levels:
                    axes[0] = plot_continuous_1d(
                        term=term,
                        gam=gam,
                        target=target,
                        data=data,
                        level=level,
                        ax=axes[0],
                        plot_kwargs={"label": level},
                        **kwargs,
                    )
                if isinstance(by_dtype, CategoricalDtype):
                    axes[0].legend()
                return axes

            return _TermPlotter(_plot_wrapper, plot_continuous_1d, target)

        case (2, AbstractTerm()) if _all_numeric(dtypes):

            def _plot_wrapper(axes, **kwargs):
                for i, level in enumerate(levels):
                    axes[i] = plot_continuous_2d(
                        term=term,
                        gam=gam,
                        target=target,
                        data=data,
                        level=level,
                        ax=axes[i],
                        **kwargs,
                    )
                    if isinstance(by_dtype, CategoricalDtype):
                        axes[i].set_title(f"Level={level}")
                return axes

            return _TermPlotter(
                _plot_wrapper,
                plot_continuous_2d,
                target,
                required_axes=len(levels),
            )

        case _:
            raise NotImplementedError(f"Did not know how to plot term {term}.")


def plot_continuous_1d(
    *,
    term: AbstractTerm,
    gam: AbstractGAM,
    target: str | None = None,
    data: pd.DataFrame | Mapping[str, pd.Series | np.ndarray] | None = None,
    eval_density: int = 100,
    level: str | None = None,
    n_standard_errors: int | float = 2,
    residuals: bool = False,
    plot_kwargs: dict[str, Any] | None = None,
    fill_between_kwargs: dict[str, Any] | None = None,
    scatter_kwargs: dict[str, Any] | None = None,
    ax: Axes | None = None,
) -> Axes:
    """Plot 1D smooth or linear terms with confidence intervals.

    !!! note

        - For terms with numeric "by" variables, the "by" variable is set to 1,
        showing the unscaled effect of the smooth.

    Args:
        term: The model term to plot. Must be a univariate term (single variable).
        gam: GAM model containing the term to plot.
        target: Name of the target variable (response variable or family
            parameter name from the model specification). If set to None, an error
            is raised when multiple predictors are present; otherwise, the sole
            available target is used.
        data: DataFrame used for plotting partial residuals and determining
            axis limits. Defaults to the data used for training.
        eval_density: Number of evaluation points along the variable range
            for plotting the smooth curve. Higher values give smoother curves
            but increase computation time. Default is 100.
        level: Must be provided for smooths with a categorical "by" variable or a
            [`FactorSmooth`][pymgcv.basis_functions.FactorSmooth] basis.
            Specifies the level to plot.
        n_standard_errors: Number of standard errors for confidence intervals.
        residuals: Whether to plot partial residuals.
        plot_kwargs: Keyword arguments passed to ``matplotlib.pyplot.plot`` for
            the main curve.
        fill_between_kwargs: Keyword arguments passed to
            `matplotlib.pyplot.fill_between` for the confidence interval band.
            Pass `{"disable": True}` to disable the confidence interval band.
        scatter_kwargs: Keyword arguments passed to `matplotlib.pyplot.scatter`
            for partial residuals (ignored if `residuals=False`).
        ax: Matplotlib Axes object to plot on. If None, uses current axes.

    Returns:
        The matplotlib Axes object with the plot.
    """
    if gam.fit_state is None:
        raise ValueError("Cannot plot before fitting the model.")

    if target is None:
        if len(gam.all_predictors) > 1:
            raise ValueError(
                "Target must be specified when multiple predictors are present.",
            )
        target = list(gam.all_predictors.keys())[0]

    data = data if data is not None else gam.fit_state.data
    data = deepcopy(data)
    term = _FactorSmoothToByInterface(term) if _is_random_wiggly(term) else term
    is_categorical_by = term.by and isinstance(data[term.by].dtype, CategoricalDtype)

    if len(term.varnames) != 1:
        raise ValueError(
            f"Expected varnames to be one continuous variable, got {term.varnames}",
        )
    if is_categorical_by and level is None:
        raise ValueError(
            "level must be provided for terms with 'by' variables, or FactorSmooths.",
        )

    if level is not None and term.by is not None:
        if isinstance(data, Mapping):
            data = {k: v[data[term.by] == level] for k, v in data.items()}
        else:
            data = data.loc[data[term.by] == level]
            assert isinstance(data, pd.DataFrame)

    x0_linspace = np.linspace(
        data[term.varnames[0]].min(),
        data[term.varnames[0]].max(),
        num=eval_density,
    )
    spaced_data = pd.DataFrame({term.varnames[0]: x0_linspace})

    if term.by is not None:
        if is_numeric_dtype(data[term.by].dtype):
            spaced_data[term.by] = 1
        else:
            spaced_data[term.by] = pd.Series(
                [level] * eval_density,
                dtype=data[term.by].dtype,
            )

    ax = plt.gca() if ax is None else ax
    plot_kwargs = {} if plot_kwargs is None else plot_kwargs
    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs
    fill_between_kwargs = {} if fill_between_kwargs is None else fill_between_kwargs
    fill_between_kwargs.setdefault("alpha", 0.2)
    scatter_kwargs.setdefault("s", 0.05 * rcParams["lines.markersize"] ** 2)

    # Matching color, particularly nice for plotting categorical by smooths on same ax
    current_color = ax._get_lines.get_next_color()  # type: ignore Can't find reasonable alternative for now
    for kwargs in (plot_kwargs, fill_between_kwargs, scatter_kwargs):
        if "c" not in kwargs and "color" not in kwargs:
            kwargs["color"] = current_color

    pred = gam.partial_effect(term, target, spaced_data, compute_se=True)

    # Add partial residuals
    if residuals and target in data and not _is_linear_functional(term, data):  # type: ignore
        partial_residuals = gam.partial_residuals(
            term,
            target=target,
            data=data,
            avoid_scaling=True,
        )

        ax.scatter(data[term.varnames[0]], partial_residuals, **scatter_kwargs)

    # Plot interval
    assert pred.se is not None
    _with_disable(ax.fill_between)(
        x0_linspace,
        pred.fit - n_standard_errors * pred.se,
        pred.fit + n_standard_errors * pred.se,
        **fill_between_kwargs,
    )

    ax.plot(x0_linspace, pred.fit, **plot_kwargs)
    ax.set_xlabel(term.varnames[0])
    ax.set_ylabel(term.label())
    return ax


def plot_continuous_2d(
    *,
    term: AbstractTerm,
    gam: AbstractGAM,
    target: str | None = None,
    data: pd.DataFrame | Mapping[str, np.ndarray | pd.Series] | None = None,
    eval_density: int = 50,
    level: str | None = None,
    contour_kwargs: dict | None = None,
    contourf_kwargs: dict | None = None,
    scatter_kwargs: dict | None = None,
    ax: Axes | None = None,
) -> Axes:
    """Plot 2D smooth surfaces as contour plots with data overlay.

    This function is essential for understanding bivariate relationships
    and interactions between two continuous variables.

    Args:
        term: The bivariate term to plot. Must have exactly two variables.
            Can be S('x1', 'x2') or T('x1', 'x2').
        gam: GAM model containing the term to plot.
        target: Name of the target variable (response variable or family
            parameter name from the model specification). If set to None, an error
            is raised when multiple predictors are present; otherwise, the sole
            available target is used.
        data: DataFrame containing the variables for determining plot range
            and showing data points. Should typically be the training data.
        eval_density: Number of evaluation points along each axis, creating
            an eval_density × eval_density grid. Higher values give smoother
            surfaces but increase computation time. Default is 50.
        level: Must be provided for smooths with a categorical "by" variable or a
            [`FactorSmooth`][pymgcv.basis_functions.FactorSmooth] basis.
            Specifies the level to plot.
        contour_kwargs: Keyword arguments passed to `matplotlib.pyplot.contour`
            for the contour lines.
        contourf_kwargs: Keyword arguments passed to `matplotlib.pyplot.contourf`
            for the filled contours.
        scatter_kwargs: Keyword arguments passed to `matplotlib.pyplot.scatter`
            for the data points overlay. Pass `{"disable": True}` to avoid plotting.
        ax: Matplotlib Axes object to plot on. If None, uses current axes.

    Returns:
        The matplotlib Axes object with the plot, allowing further customization.

    Raises:
        ValueError: If the term doesn't have exactly two variables.
    """
    if gam.fit_state is None:
        raise ValueError("Cannot plot before fitting the model.")

    if target is None:
        if len(gam.all_predictors) > 1:
            raise ValueError(
                "Target must be specified when multiple predictors are present.",
            )
        target = list(gam.all_predictors.keys())[0]

    data = data if data is not None else gam.fit_state.data
    data = deepcopy(data)
    term = _FactorSmoothToByInterface(term) if _is_random_wiggly(term) else term
    is_categorical_by = term.by and isinstance(data[term.by].dtype, CategoricalDtype)

    if len(term.varnames) != 2:
        raise ValueError(
            f"Expected varnames to be one continuous variable, got {term.varnames}",
        )

    if is_categorical_by and level is None:
        raise ValueError(
            "level must be provided for terms with 'by' variables, or FactorSmooths.",
        )

    if level is not None and term.by is not None:
        if isinstance(data, Mapping):
            data = {k: v[data[term.by] == level] for k, v in data.items()}
        else:
            data = data.loc[data[term.by] == level]
            assert isinstance(data, pd.DataFrame)

    x0_lims = (data[term.varnames[0]].min(), data[term.varnames[0]].max())
    x1_lims = (data[term.varnames[1]].min(), data[term.varnames[1]].max())
    x0_mesh, x1_mesh = np.meshgrid(
        np.linspace(*x0_lims, eval_density),
        np.linspace(*x1_lims, eval_density),
    )
    spaced_data = pd.DataFrame(
        {term.varnames[0]: x0_mesh.ravel(), term.varnames[1]: x1_mesh.ravel()},
    )
    if term.by is not None:
        if is_numeric_dtype(data[term.by].dtype):
            spaced_data[term.by] = 1
        else:
            spaced_data[term.by] = pd.Series(
                [level] * eval_density**2,
                dtype=data[term.by].dtype,
            )

    ax = plt.gca() if ax is None else ax
    contour_kwargs = {} if contour_kwargs is None else contour_kwargs
    contourf_kwargs = {} if contourf_kwargs is None else contourf_kwargs
    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs

    contour_kwargs.setdefault("levels", 14)
    contourf_kwargs.setdefault("levels", 14)
    contourf_kwargs.setdefault("alpha", 0.8)
    scatter_kwargs.setdefault("color", "black")
    scatter_kwargs.setdefault("s", 0.05 * rcParams["lines.markersize"] ** 2)
    scatter_kwargs.setdefault("zorder", 2)  # Ensures above contours

    pred = gam.partial_effect(
        term,
        target,
        data=spaced_data,
    )

    mesh = ax.contourf(
        x0_mesh,
        x1_mesh,
        pred.reshape(x0_mesh.shape),
        **contourf_kwargs,
    )
    _with_disable(ax.contour)(
        x0_mesh,
        x1_mesh,
        pred.reshape(x0_mesh.shape),
        **contour_kwargs,
    )
    color_bar = ax.figure.colorbar(mesh, ax=ax, pad=0)
    color_bar.set_label(term.label())
    _with_disable(ax.scatter)(
        np.asarray(data[term.varnames[0]]).ravel(),  # Ravel for linear functional terms
        np.asarray(data[term.varnames[1]]).ravel(),
        **scatter_kwargs,
    )
    ax.set_xlabel(term.varnames[0])
    ax.set_ylabel(term.varnames[1])
    return ax


def plot_categorical(
    *,
    term: L,
    gam: AbstractGAM,
    target: str | None = None,
    data: pd.DataFrame | Mapping[str, pd.Series | np.ndarray] | None = None,
    residuals: bool = False,
    n_standard_errors: int | float = 2,
    errorbar_kwargs: dict[str, Any] | None = None,
    scatter_kwargs: dict[str, Any] | None = None,
    ax: Axes | None = None,
) -> Axes:
    """Plot categorical terms with error bars and partial residuals.

    Creates a plot showing:

    - The estimated effect of each category level as points.
    - Error bars representing confidence intervals.
    - Partial residuals as jittered scatter points.

    Args:
        term: The categorical term to plot. Must be a L term with a single
            categorical variable.
        gam: GAM model containing the term to plot.
        target: Name of the target variable (response variable or family
            parameter name from the model specification). If set to None, an error
            is raised when multiple predictors are present; otherwise, the sole
            available target is used.
        data: DataFrame (or dictionary) containing the categorical variable and
            response variable.
        residuals: Whether to plot partial residuals (jittered on x-axis).
        n_standard_errors: Number of standard errors for confidence intervals.
        errorbar_kwargs: Keyword arguments passed to `matplotlib.pyplot.errorbar`.
        scatter_kwargs: Keyword arguments passed to `matplotlib.pyplot.scatter`.
        ax: Matplotlib Axes object to plot on. If None, uses current axes.

    """
    if gam.fit_state is None:
        raise RuntimeError("The model must be fitted before plotting.")

    if target is None:
        if len(gam.all_predictors) > 1:
            raise ValueError(
                "Target must be specified when multiple predictors are present.",
            )
        target = list(gam.all_predictors.keys())[0]

    data = gam.fit_state.data if data is None else data

    errorbar_kwargs = {} if errorbar_kwargs is None else errorbar_kwargs
    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs
    scatter_kwargs.setdefault("s", 0.05 * rcParams["lines.markersize"] ** 2)
    errorbar_kwargs.setdefault("capsize", 10)
    errorbar_kwargs.setdefault("fmt", ".")

    ax = plt.gca() if ax is None else ax
    vals = data[term.varnames[0]]

    if not isinstance(vals.dtype, CategoricalDtype):
        raise TypeError("The variable must be categorical in the data.")
    assert isinstance(vals, pd.Series)

    levels = pd.Series(
        vals.cat.categories,
        dtype=vals.dtype,
        name=term.varnames[0],
    )

    if residuals and target in data:
        partial_residuals = gam.partial_residuals(term, target, data)

        jitter = np.random.uniform(-0.25, 0.25, size=data_len(data))
        scatter_kwargs.setdefault("alpha", 0.2)

        ax.scatter(
            vals.cat.codes + jitter,
            partial_residuals,
            **scatter_kwargs,
        )

    ax.set_xticks(ticks=levels.cat.codes, labels=levels)

    pred = gam.partial_effect(
        term=term,
        target=target,
        data=pd.DataFrame(levels),
        compute_se=True,
    )

    assert pred.se is not None

    ax.errorbar(
        x=levels.cat.codes,
        y=pred.fit,
        yerr=n_standard_errors * pred.se,
        **errorbar_kwargs,
    )
    ax.set_xlabel(term.varnames[0])
    ax.set_ylabel(term.label())
    return ax


def plot_random_effect(
    *,
    term: S,
    gam: AbstractGAM,
    target: str | None = None,
    confidence_interval_level: float = 0.95,
    axline_kwargs: dict[str, Any] | None = None,
    scatter_kwargs: dict[str, Any] | None = None,
    fill_between_kwargs: dict[str, Any] | None = None,
    ax: Axes | None = None,
) -> Axes:
    """A QQ-like-plot for random effect terms.

    This function plots the estimated random effects against Gaussian quantiles and
    includes a confidence envelope to assess whether the random effects follow a
    normal distribution, as assumed by the model.

    Args:
        term: The random effect term to plot. Must be a smooth term with a
            [`RandomEffect`][pymgcv.basis_functions.RandomEffect] basis function.
        gam: The fitted GAM model containing the random effect.
        target: The target variable to plot when multiple predictors are present.
            If None and only one predictor exists, that predictor is used.
        confidence_interval_level: The confidence level for the confidence envelope.
        axline_kwargs: Keyword arguments passed to `matplotlib.axes.Axes.axline` for
            the reference line.
        scatter_kwargs: Keyword arguments passed to `matplotlib.axes.Axes.scatter` for
            the random effect points.
        fill_between_kwargs: Keyword arguments passed to
            `matplotlib.axes.Axes.fill_between` for the confidence envelope.
        ax: Matplotlib axes to use for the plot. If None, uses the current axes.

    Returns:
        The matplotlib axes object.

    !!! note
        The confidence interval calculation is based on the formula from:
        "Worm plot: a simple diagnostic device for modelling growth reference curves"
        (page 6). The random effects are constrained to be centered, so the reference
        line passes through (0, 0).


    """
    if gam.fit_state is None:
        raise RuntimeError("The model must be fitted before plotting.")

    if not isinstance(term.bs, RandomEffect):
        raise TypeError("Term is not a random effect term.")

    if target is None:
        if len(gam.all_predictors) > 1:
            raise ValueError(
                "Target must be specified when multiple predictors are present.",
            )
        target = list(gam.all_predictors.keys())[0]

    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs
    scatter_kwargs.setdefault("s", 0.05 * rcParams["lines.markersize"] ** 2)

    fill_between_kwargs = {} if fill_between_kwargs is None else fill_between_kwargs
    current_color = ax._get_lines.get_next_color()  # type: ignore Can't find reasonable alternative for now
    for kwargs in (scatter_kwargs, fill_between_kwargs):
        if "c" not in kwargs and "color" not in kwargs:
            kwargs["color"] = current_color

    axline_kwargs = {} if axline_kwargs is None else axline_kwargs

    if "c" not in axline_kwargs and "color" not in axline_kwargs:
        axline_kwargs["color"] = "gray"

    axline_kwargs.setdefault("linestyle", "--")
    ax = plt.gca() if ax is None else ax

    # TODO remove data argument?
    levels = np.unique(gam.fit_state.data[term.varnames[0]])
    pred = gam.partial_effect(
        term=term,
        target=target,
        data=pd.DataFrame(pd.Series(levels, name=term.varnames[0], dtype="category")),
    )
    order = np.argsort(pred)
    x = rstats.qnorm(rstats.ppoints(len(levels)))  # Gaussian quantiles
    ax.scatter(
        x,
        pred[order],
        **scatter_kwargs,
    )

    # Add CI lines. Confidence interval based on formula in
    # "Worm plot: a simple diagnostic device for modelling growth reference curves"
    # page 6. We need to multiply them by sd(.dat$y) because we are not normalizing the
    # random effects.
    alpha = (1 - confidence_interval_level) / 2
    n = len(levels)
    p = to_py(rstats.ppoints(n))
    interval = (
        np.std(pred)
        * to_py(rstats.qnorm(alpha))  # e.g. 1.96
        * np.sqrt(p * (1 - p) / n)
        / to_py(rstats.dnorm(x))
    )
    ref_y = x * np.std(pred)
    _with_disable(ax.fill_between)(
        x,
        ref_y - interval,
        ref_y + interval,
        alpha=0.2,
        **fill_between_kwargs,
    )
    # RE constrained to be centered, so plot reference through (0, 0)
    _with_disable(ax.axline)(
        (0, 0),
        slope=np.std(pred).item(),
        **axline_kwargs,
    )  # TODO use axline in qq too?
    ax.set_xlabel("Gaussian Quantiles")
    ax.set_ylabel(f"{term.varnames[0]} effect")  # TODO target?
    return ax


def plot_qq(
    gam: AbstractGAM,
    *,
    qq_fun: Callable[[AbstractGAM], QQResult] = qq_simulate,
    scatter_kwargs: dict | None = None,
    fill_between_kwargs: dict | None = None,
    plot_kwargs: dict | None = None,
    ax: Axes | None = None,
) -> Axes:
    """A Q-Q plot of deviance residuals.

    Args:
        gam: The fitted GAM model.
        qq_fun: A function taking only the GAM model, and returning a `QQResult`
            object storing the theoretical residuals, residuals, and the confidence
            interval. Defaults to [`qq_simulate`][pymgcv.qq.qq_simulate], which is the
            most widely supported method only requiring the family to provide a
            sampling function. [`qq_cdf`][pymgcv.qq.qq_cdf] can be used for families
            providing a cdf method, which transforms the data to a uniform
            distribution

            !!! warning

                `qq_cdf` transforms all data to [0, 1] uniform. This may make it hard to
                spot deviations in tail behaviour, e.g. due to a few extreme outliers.

        scatter_kwargs: Key word arguments passed to `matplotlib.pyplot.scatter`.
        fill_between_kwargs: Key word arguments passed to
            `matplotlib.pyplot.fill_between`, for plotting the confidence interval.
        plot_kwargs: Key word arguments passed to `matplotlib.pyplot.plot` for
            plotting the reference line. Pass {"disable": True} to avoid plotting.
        ax: Matplotlib axes to use for the plot.


    !!! note

        We can change e.g. the confidence level and number of simulations using
        partial application:
        ```python
        from pymgcv.qq import qq_simulate
        from functools import partial

        qq_fun = partial(qq_simulate, level=0.95, n_sim=10)
        # plot_qq(..., qq_fun=qq_fun)
        ```

    Returns:
        The matplotlib axes object.

    !!! example

        As an example, we will create a heavy tailed response variable,
        and fit a [`Gaussian`][pymgcv.families.Gaussian] model, and a
        [`Scat`][pymgcv.families.Scat] model.

        ```python
        import matplotlib.pyplot as plt
        import numpy as np
        import pandas as pd

        from pymgcv.families import Gaussian, Scat
        from pymgcv.gam import GAM
        from pymgcv.plot import plot_qq
        from pymgcv.terms import S

        rng = np.random.default_rng(1)
        n = 1000
        x = np.linspace(0, 1, n)
        y = np.sin(2 * np.pi * x) + rng.standard_t(df=3, size=n)  # Heavy-tailed
        data = pd.DataFrame({"x": x, "y": y})

        models = [
            GAM({"y": S("x")}, family=Gaussian()),
            GAM({"y": S("x")}, family=Scat()),  # Better for heavy-tailed data
        ]

        fig, axes = plt.subplots(ncols=2)

        for model, ax in zip(models, axes, strict=False):
            model.fit(data)
            plot_qq(model, ax=ax)
            ax.set_title(model.family.__class__.__name__)
            ax.set_box_aspect(1)

        # fig.show()  # Uncomment to display the figure
        ```
    """
    if gam.fit_state is None:
        raise RuntimeError("The model must be fitted before plotting.")

    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs
    scatter_kwargs.setdefault("s", 0.2 * rcParams["lines.markersize"] ** 2)
    fill_between_kwargs = {} if fill_between_kwargs is None else fill_between_kwargs
    fill_between_kwargs.setdefault("alpha", 0.4)

    if "c" not in scatter_kwargs and "color" not in scatter_kwargs:
        scatter_kwargs["color"] = "black"

    plot_kwargs = {} if plot_kwargs is None else plot_kwargs

    if "c" not in plot_kwargs and "color" not in plot_kwargs:
        plot_kwargs["color"] = "gray"
    plot_kwargs.setdefault("linestyle", "--")
    ax = plt.gca() if ax is None else ax
    qq_data = qq_fun(gam)

    _with_disable(ax.fill_between)(
        qq_data.theoretical,
        *qq_data.interval,
        **fill_between_kwargs,
    )

    ax.scatter(qq_data.theoretical, qq_data.residuals, **scatter_kwargs)
    ax.set_xlabel("Theoretical Quantiles")
    ax.set_ylabel("Residuals")

    min_val = min(ax.get_xlim()[0], ax.get_ylim()[0])
    max_val = max(ax.get_xlim()[1], ax.get_ylim()[1])
    _with_disable(ax.plot)([min_val, max_val], [min_val, max_val], **plot_kwargs)
    return ax


def plot_residuals_vs_linear_predictor(
    gam: AbstractGAM,
    type: Literal[
        "deviance",
        "pearson",
        "scaled.pearson",
        "working",
        "response",
    ] = "deviance",
    target: str | None = None,
    ax: Axes | None = None,
    scatter_kwargs: dict[str, Any] | None = None,
):
    """Plot the residuals against the linear predictor.

    Args:
        gam: The fitted GAM model.
        type: The type of residuals to plot.
        target: The target variable to plot residuals for.
        ax: The axes to plot on.
        scatter_kwargs: Keyword arguments to pass to the scatter plot.
    """
    if len(gam.predictors) > 1:
        raise NotImplementedError(
            "Multivariate response families are not supported for "
            "plot_residuals_vs_linear_predictor.",
        )
    scatter_kwargs = {} if scatter_kwargs is None else scatter_kwargs
    scatter_kwargs.setdefault("s", 0.05 * rcParams["lines.markersize"] ** 2)

    ax = plt.gca() if ax is None else ax
    residuals = gam.residuals(type)
    if target is None:
        if len(gam.all_predictors) > 1:
            raise ValueError(
                "Target must be specified when multiple predictors are present.",
            )
        target = list(gam.all_predictors.keys())[0]
    predictions = gam.predict()[target]
    ax.scatter(predictions, residuals, **scatter_kwargs)
    ax.set_xlabel("Linear predictor")
    ax.set_ylabel("Residuals")
    return ax


def hexbin_residuals(
    residuals: np.ndarray,
    var1: str,
    var2: str,
    data: pd.DataFrame | Mapping[str, np.ndarray | pd.Series],
    *,
    gridsize: int = 25,
    max_val: int | float | None = None,
    ax: Axes | None = None,
    **kwargs,
):
    """Hexbin plot for visualising residuals as function of two variables.

    Useful e.g. for assessing if interactions are might be required. This
    is a thin wrapper around `matplotlib.pyplot.hexbin`, with better defaults
    for plotting residuals (e.g. uses a symmetric color scale).

    The default reduction function is `np.sum(res) / np.sqrt(len(res))`,
    which has constant variance w.r.t. the number of points.

    Args:
        residuals: Residuals to plot.
        var1: Name of the first variable.
        var2: Name of the second variable.
        data: The data (containing ``var1`` and ``var2``).
        gridsize: The number of hexagons in the x-direction. The y direction is chosen
            such that the hexagons are approximately regular.
        max_val: Maximum and minimum value for the symmetric color scale. Defaults to
            the maximum absolute value of the residuals.
        ax: Axes to plot on. If None, the current axes are used.
        **kwargs: Additional keyword arguments passed to matplotlib hexbin.

    !!! example

        ```python
        import numpy as np
        from pymgcv.plot import hexbin_residuals
        import matplotlib.pyplot as plt

        rng = np.random.default_rng(1)

        fig, ax = plt.subplots()
        residuals = rng.normal(size=500)  # or gam.residuals()
        data = {
            "x0": rng.normal(size=residuals.shape),
            "x1": rng.normal(size=residuals.shape),
            }

        hexbin_residuals(residuals, "x0", "x1", data=data, ax=ax)
        ```
    """
    max_color = np.max(np.abs(residuals)) if max_val is None else max_val
    kwargs.setdefault("cmap", "coolwarm")
    kwargs.setdefault("reduce_C_function", lambda x: np.sum(x) / np.sqrt(len(x)))
    kwargs.setdefault("vmin", -max_color)
    kwargs.setdefault("vmax", max_color)
    ax = plt.gca() if ax is None else ax
    ax.hexbin(
        data[var1],
        data[var2],
        gridsize=gridsize,
        C=residuals,
        **kwargs,
    )
    ax.set_xlabel(var1)
    ax.set_ylabel(var2)
    return ax


def _with_disable(plot_func):
    """Wraps a plot function to easily disable with disable=True."""

    def wrapper(*args, disable=False, **kwargs):
        if disable:
            return None
        return plot_func(*args, **kwargs)

    return wrapper


def _is_random_wiggly(term: AbstractTerm) -> TypeGuard[T | S]:
    if isinstance(term, S | T):
        return isinstance(term.bs, FactorSmooth)
    return False


def _is_linear_functional(
    term: AbstractTerm,
    data: pd.DataFrame | dict[str, pd.Series | np.ndarray],
) -> bool:
    return any(np.asarray(data[k]).ndim > 1 for k in term.varnames) and isinstance(
        term,
        S | T,
    )



import numpy as np
import pandas as pd
import rpy2.robjects as ro

from pymgcv.rpy_utils import data_to_rdf


def test_data_to_rdf():
    d = pd.DataFrame({"a": np.array([1, 2, 3]), "b": np.array([4, 5, 6])})
    df = data_to_rdf(d, include="all")
    assert df.nrow == 3
    assert df.ncol == 2
    assert list(df.rx2("a")) == [1, 2, 3]
    assert list(df.rx2("b")) == [4, 5, 6]


def test_data_to_rdf_with_matrix():
    d = {"a": np.array([1, 2, 3]), "b": np.ones((3, 4))}
    df = data_to_rdf(d, include="all")
    assert df.nrow == 3
    assert df.ncol == 2
    assert df.rx2["b"].ncol == 4


def test_data_to_rdf_categorical_factors():
    data = pd.DataFrame(
        {
            "y": np.arange(3),
            "x": pd.Categorical(
                ["green", "green", "blue"],
                categories=["red", "green", "blue"],
            ),
        },
    )

    rdf = data_to_rdf(data, include="all")
    factor = rdf.rx2("x")
    assert isinstance(factor, ro.vectors.FactorVector)
    assert factor.nlevels == 3

    rdf = data_to_rdf(pd.DataFrame(data), include="all")
    factor = rdf.rx2("x")
    assert isinstance(factor, ro.vectors.FactorVector)
    assert factor.nlevels == 3



import pandas as pd

from pymgcv.utils import load_rdata_dataframe_from_url


def test_load_rdata_dataframe_from_url():
    url = "https://github.com/mfasiolo/testGam/raw/master/data/Larynx.rda"
    data = load_rdata_dataframe_from_url(url)
    assert isinstance(data, pd.DataFrame)



"""We centralise the imports here to avoid overhead from repeated rpy2 imports."""

from rpy2.robjects.packages import importr

rbase = importr("base")
rmgcv = importr("mgcv")
rstats = importr("stats")
rutils = importr("utils")



"""Formula utilities."""

from dataclasses import dataclass
from typing import Any

from rpy2 import robjects as ro

from pymgcv.rlibs import rbase
from pymgcv.rpy_utils import to_rpy


@dataclass
class _Var:
    name: str


def _to_r_constructor_string(arg: Any) -> str:
    """Converts an object to R string representation.

    _Var acts as a placeholder for a variable name in R.

    Args:
        arg: The object to convert. If not already an RObject, it is first passed to
            `to_rpy`.
    """
    if isinstance(arg, _Var):
        return arg.name

    if not isinstance(arg, ro.RObject):
        arg = to_rpy(arg)

    connection = rbase.textConnection("__r_obj_str", "w")
    rbase.dput(arg, file=connection)
    rbase.close(connection)
    result = ro.r["__r_obj_str"]
    assert len(result) == 1  # type: ignore
    return result[0]  # type: ignore



"""Custom types for pymgcv."""

from dataclasses import dataclass
from typing import Generic, TypeVar

import numpy as np
import pandas as pd

S = TypeVar("S", np.ndarray, pd.DataFrame)


@dataclass
class FitAndSE(Generic[S]):
    """Container for predictions or partial effects with standard errors.

    Attributes:
        fit: Predicted values or partial effect.
        se: Standard errors of the predictions.
    """

    fit: S
    se: S



from collections.abc import Mapping

import numpy as np
import pandas as pd
import rpy2.rinterface_lib.callbacks as rcb
import rpy2.robjects as ro

from pymgcv.rlibs import rbase, rutils
from pymgcv.rpy_utils import to_py


def load_rdata_dataframe_from_url(url: str) -> pd.DataFrame:
    """Load an RData (.rda) file from URL expecting a single dataframe, without R console output."""
    # Backup the original R console print callback
    original_consolewrite_print = rcb.consolewrite_print

    # Override with a dummy function to suppress output
    rcb.consolewrite_print = lambda x: None  # ignore any output
    rcb.consolewrite_warnerror = lambda x: None  # also ignore warnings if needed

    try:
        temp_file = rbase.tempfile(fileext=".RData")
        rutils.download_file(url, destfile=temp_file, mode="wb")
        loaded_names = rbase.load(temp_file)
        loaded_names_py = list(loaded_names)

        if len(loaded_names_py) != 1:
            raise ValueError(
                f"Expected one object, got {len(loaded_names_py)}: {loaded_names_py}",
            )

        data_r = ro.r[loaded_names_py[0]]
        data = to_py(data_r)
        rbase.file_remove(temp_file)

        if not isinstance(data, pd.DataFrame):
            raise TypeError(f"Expected a dataframe, got {type(data)}")

        return data

    finally:
        # Restore original R console print callback
        rcb.consolewrite_print = original_consolewrite_print
        rcb.consolewrite_warnerror = rcb.consolewrite_print


def get_data(name: str):
    """Get built-in R dataset.

    Currently assumes that the dataset is a dataframe.
    """
    with ro.local_context() as lc:
        rutils.data(ro.rl(name), envir=lc)
        return to_py(lc[name])


def data_len(data: pd.DataFrame | Mapping[str, pd.Series | np.ndarray]) -> int:
    """Get the length of the data.

    If the data is a dictionary, returns the maximum value of the shape along axis 0.
    """
    if isinstance(data, pd.DataFrame):
        return len(data)
    return max([d.shape[0] for d in data.values()])



"""Smoke tests for plotting functions, simply check that they run without errors.

To see the plots created during testing, replace plt.close("all") with plt.show()
"""

import matplotlib.pyplot as plt
import pytest

from pymgcv.gam import BAM, GAM, AbstractGAM
from pymgcv.plot import (
    plot_categorical,
    plot_continuous_1d,
    plot_continuous_2d,
    plot_gam,
    plot_qq,
    plot_residuals_vs_linear_predictor,
)

from . import gam_test_cases as tc


def get_cases_1d_continuous(gam_type: type[AbstractGAM]):
    cases = [
        (tc.linear_gam, {}),
        (tc.smooth_1d_gam, {"residuals": True}),
        (tc.smooth_1d_by_numeric_gam, {}),
        (tc.smooth_1d_random_wiggly_curve_gam, {"level": "A"}),
        (tc.smooth_1d_by_categorical_gam, {"level": "A"}),
        (tc.linear_functional_smooth_1d_gam, {}),
    ]
    return {
        f"{gam_type.__name__} - {f.__name__}": (f(gam_type), kwargs)
        for f, kwargs in cases
    }


cases_1d_continuous = get_cases_1d_continuous(GAM) | get_cases_1d_continuous(BAM)


@pytest.mark.parametrize(
    ("test_case", "kwargs"),
    cases_1d_continuous.values(),
    ids=cases_1d_continuous.keys(),
)
def test_plot_continuous_1d(test_case: tc.GAMTestCase, kwargs: dict):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    term = list(gam.all_predictors.values())[0][0]
    plot_continuous_1d(**kwargs, gam=gam, term=term, data=test_case.data)
    plt.close("all")


def get_cases_2d_continuous(gam_type: type[AbstractGAM]):
    test_cases_1d_continuous = [
        (tc.smooth_2d_gam, {}),
        (tc.tensor_2d_gam, {}),
        (tc.tensor_2d_by_numeric_gam, {}),
        (tc.tensor_2d_by_categorical_gam, {"level": "A"}),
        (tc.tensor_2d_random_wiggly_curve_gam, {"level": "A"}),
        (tc.linear_functional_tensor_2d_gam, {}),
    ]
    return {
        f"{gam_type.__name__} - {f.__name__}": (f(gam_type), kwargs)
        for f, kwargs in test_cases_1d_continuous
    }


cases_2d_continuous = get_cases_2d_continuous(GAM) | get_cases_2d_continuous(BAM)


@pytest.mark.parametrize(
    ("test_case", "kwargs"),
    cases_2d_continuous.values(),
    ids=cases_2d_continuous.keys(),
)
def test_plot_continuous_2d(test_case: tc.GAMTestCase, kwargs: dict):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    term = list(gam.all_predictors.values())[0][0]
    plot_continuous_2d(**kwargs, gam=gam, term=term, data=test_case.data)
    plt.close("all")


@pytest.mark.parametrize(
    "gam_type",
    [GAM, BAM],
)
def test_plot_categorical(gam_type: type[AbstractGAM]):
    test_case = tc.categorical_linear_gam(gam_type)
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    term = list(gam.all_predictors.values())[0][0]
    plot_categorical(target="y", gam=gam, term=term, data=test_case.data)
    plt.close("all")


all_test_cases = (
    cases_1d_continuous
    | cases_2d_continuous
    | {"categorical_linear": (tc.categorical_linear_gam(GAM), {"target": "y"})}
)

all_gam_test_cases = tc.get_test_cases()


@pytest.mark.parametrize(
    "test_case",
    all_gam_test_cases.values(),
    ids=all_gam_test_cases.keys(),
)
def test_plot_gam(test_case: tc.GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    try:
        plot_gam(gam=gam, ncols=1)  # scatter=True fails for mvn + gaulss
    except (ValueError, NotImplementedError) as e:
        if "plot any" in str(e):
            pass
        else:
            raise
    plt.close("all")


@pytest.mark.parametrize(
    "test_case",
    all_gam_test_cases.values(),
    ids=all_gam_test_cases.keys(),
)
def test_plot_qq(test_case: tc.GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    try:
        plot_qq(gam=gam)
    except NotImplementedError as e:
        if "Multivariate response" in str(e):  # e.g. mvn and gaulss
            pass
        else:
            raise
    except TypeError as e:
        if "Family must support CDF method" in str(e):
            pass
        else:
            raise

    plt.close("all")


@pytest.mark.parametrize(
    "test_case",
    all_gam_test_cases.values(),
    ids=all_gam_test_cases.keys(),
)
def test_plot_residuals_vs_linear_predictor(test_case: tc.GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    target = list(gam.all_predictors.keys())[0]
    try:
        plot_residuals_vs_linear_predictor(gam, target=target)
    except NotImplementedError as e:
        if "Multivariate response" in str(e):  # e.g. mvn and gaulss
            pass
        else:
            raise

    plt.close("all")



"""Families supported by pymgcv."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Literal

import numpy as np
import rpy2.robjects as ro

from pymgcv.rlibs import rbase, rmgcv, rstats
from pymgcv.rpy_utils import is_null, to_py, to_rpy

# For qq_simulate we need sample, and for qq_cdf we need cdf implemented.


class AbstractFamily(ABC):
    """Provides default implmentations for distribution methods.

    This applies mgcv `fix.family.qf` for the quantile function, and
    `fix.family.rd` for the sampling function.
    """

    rfamily: ro.ListVector

    def link(self, x: np.ndarray) -> np.ndarray:
        """Compute the link function."""
        result = to_py(self.rfamily.rx2["linkfun"](to_rpy(x)))
        assert isinstance(result, np.ndarray)
        return result

    def inverse_link(self, x: np.ndarray) -> np.ndarray:
        """Compute the inverse link function."""
        result = to_py(self.rfamily.rx2["linkinv"](to_rpy(x)))
        assert isinstance(result, np.ndarray)
        return result

    def dmu_deta(self, x: np.ndarray) -> np.ndarray:
        """Compute the derivative dmu/deta of the link function."""
        result = to_py(self.rfamily.rx2["mu.eta"](to_rpy(x)))
        assert isinstance(result, np.ndarray)
        return result

    def sample(
        self,
        mu: int | float | np.ndarray,
        wt: int | float | np.ndarray | None = None,
        scale: int | float | np.ndarray | None = None,
    ):
        """Sample the family distributions (R family rd method)."""
        sample_fn = rmgcv.fix_family_rd(self.rfamily).rx2["rd"]
        if is_null(sample_fn):
            raise NotImplementedError(
                f"Sample function not available for family {self.__name__}.",
            )

        kwargs = {"mu": mu, "wt": wt, "scale": scale}
        # R won't broadcast length 1 arrays
        kwargs = {
            k: to_rpy(v) if v.size != 1 else v.item()
            for k, v in kwargs.items()
            if v is not None
        }
        return to_py(sample_fn(**kwargs))


class SupportsCDF(ABC):
    """Mixin for families supporting cumulative distribution functions."""

    @abstractmethod
    def cdf(
        self,
        x: np.ndarray,
        *,
        mu: np.ndarray,
        wt: np.ndarray,
        scale: np.ndarray,
    ) -> np.ndarray:
        """Cumulative distribution function."""
        pass


@dataclass
class Gaussian(AbstractFamily, SupportsCDF):
    """Gaussian family with specified link function.

    Args:
        link: The link function.
    """

    def __init__(self, link: Literal["identity", "log", "inverse"] = "identity"):
        self.rfamily = rstats.gaussian(link=link)

    def cdf(
        self,
        x: np.ndarray,
        *,
        mu: np.ndarray,
        wt: np.ndarray,
        scale: np.ndarray,
    ) -> np.ndarray:
        """Gaussian CDF."""
        sd = np.sqrt(scale / wt)
        x, mu, sd = (to_rpy(arr) for arr in np.broadcast_arrays(x, mu, sd))
        return to_py(rstats.pnorm(x, mean=mu, sd=sd))


# TODO another case where matrix inputs need to be supported.
# TODO document response form
@dataclass
class Binomial(AbstractFamily, SupportsCDF):
    """Binomial family with specified link function.

    Args:
        link: The link function. "logit", "probit" and "cauchit", correspond to
            logistic, normal and Cauchy CDFs respectively. "cloglog" is the
            complementary log-log.
    """

    def __init__(
        self,
        link: Literal["logit", "probit", "cauchit", "log", "cloglog"] = "logit",
    ):
        self.rfamily = rstats.binomial(link=link)

    def cdf(
        self,
        x: np.ndarray,
        *,
        mu: np.ndarray,
        wt: np.ndarray,
        scale: np.ndarray,
    ):
        """Binomial CDF, scale is ignored."""
        x, mu, wt = (to_rpy(arr) for arr in np.broadcast_arrays(x, mu, wt))
        return to_py(rstats.pbinom(x * (wt + rbase.as_numeric(wt == 0)), wt, mu))


class Gamma(AbstractFamily, SupportsCDF):
    """Gamma family with specified link function.

    Args:
        link: The link function for the Gamma family.
    """

    def __init__(self, link: Literal["inverse", "identity", "log"] = "inverse"):
        self.rfamily = rstats.Gamma(link=link)

    def cdf(
        self,
        x: np.ndarray,
        *,
        mu: np.ndarray,
        wt: np.ndarray,
        scale: np.ndarray,
    ):
        """Gamma CDF, wt is ignored."""
        x, mu, scale = [to_rpy(arr) for arr in np.broadcast_arrays(x, mu, scale)]
        return to_py(
            rstats.pgamma(to_rpy(x), shape=to_rpy(1 / scale), scale=to_rpy(mu * scale)),
        )


class InverseGaussian(AbstractFamily):
    """Inverse Gaussian family with specified link function.

    Args:
        link: The link function for the inverse Gaussian family.
    """

    def __init__(
        self,
        link: Literal["1/mu^2", "inverse", "identity", "log"] = "1/mu^2",
    ):
        self.rfamily = rstats.inverse_gaussian(link=link)


class Poisson(AbstractFamily, SupportsCDF):
    """Poisson family with specified link function.

    Args:
        link: The link function for the Poisson family.
    """

    def __init__(self, link: Literal["log", "identity", "sqrt"] = "log"):
        self.rfamily = rstats.poisson(link=link)

    def cdf(
        self,
        x: np.ndarray,
        *,
        mu: np.ndarray,
        wt: np.ndarray | None = None,
        scale: np.ndarray | None = None,
    ):
        """Cumulative distribution function."""
        return to_py(rstats.ppois(to_rpy(x), to_rpy(mu)))


# TODO These won't support sampling
class Quasi(AbstractFamily):
    """Quasi family with specified link and variance functions.

    Args:
        link: The link function for the quasi family. Valid options are:
        variance: The variance function for the quasi family.
    """

    def __init__(
        self,
        link: Literal[
            "logit",
            "probit",
            "cloglog",
            "identity",
            "inverse",
            "log",
            "1/mu^2",
            "sqrt",
        ] = "identity",
        variance: Literal["constant", "mu(1-mu)", "mu", "mu^2", "mu^3"] = "constant",
    ):
        self.rfamily = rstats.quasi(link=link, variance=variance)


class QuasiBinomial(AbstractFamily):
    """Quasi-binomial family with specified link function.

    Args:
        link: The link function for the quasi-binomial family.
    """

    def __init__(
        self,
        link: Literal["logit", "probit", "cauchit", "log", "cloglog"] = "logit",
    ):
        self.rfamily = rstats.quasibinomial(link=link)


class QuasiPoisson(AbstractFamily):
    """Quasi-Poisson family with specified link function.

    Args:
        link: The link function for the quasi-Poisson family.
    """

    def __init__(self, link: Literal["log", "identity", "sqrt"] = "log"):
        self.rfamily = rstats.quasipoisson(link=link)


class Betar(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class CNorm(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class CLog(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class CPois(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class NegBin(AbstractFamily):
    r"""Negative binomial family.

    Args:
        theta: The positive parameter such that
            $\text{var}(y) = \mu + \mu^2/\theta$, where $\mu = \mathbb{E}[y]$.
        link: The link function to use.
        theta_fixed: Whether to treat theta as fixed or estimated. If estimated,
            then theta is the starting value.
    """

    def __init__(
        self,
        theta: float | int | None = None,
        link: Literal["log", "identity", "sqrt"] = "log",
        *,
        theta_fixed: bool = False,
    ):
        # For now this just uses nb family (not negbin)
        if theta_fixed and theta is None:
            raise ValueError("Theta must be specified if fixed.")

        if theta is not None:
            theta = theta if theta_fixed else -theta  # mgcv convention

        self.rfamily = rmgcv.nb(theta=ro.NULL if theta is None else theta, link=link)


class OCat(AbstractFamily):
    """Ordered categorical family.

    For performing regression with ordered categorical data.

    Args:
        num_categories: The number of categories.
    """

    def __init__(self, num_categories: int):
        self.rfamily = rmgcv.ocat(R=num_categories)


class Scat(AbstractFamily):
    r"""Scaled t family for heavy tailed data.

    Args:
        link: The link function to use.
        min_df: The minimum degrees of freedom. Must be >2 to avoid infinite
            response variance.
        theta: The parameters to be estimated $\nu = b + \exp(\theta_1)$
            (where $b$ is `min_df`) and $\sigma = \exp(\theta_2)$. If supplied
            and both positive, then taken to be fixed values of $\nu$ and
            $\sigma$. If any negative, then absolute values taken as starting
            values.
        theta_fixed: If theta is provided, controls whether to treat theta as fixed
            or estimated. If estimated, then theta is the starting value.
    """

    def __init__(
        self,
        link: Literal["identity", "log", "inverse"] = "identity",
        min_df: float | int = 3,
        theta: np.ndarray | None = None,
        *,
        theta_fixed: bool = False,
    ):
        if theta is not None and not theta_fixed:
            theta = -theta  # mgcv convention.
        self.rfamily = rmgcv.scat(link=link, min_df=min_df)


class Tweedie(AbstractFamily):
    r"""Tweedie family with fixed power.

    Args:
        p: The variance of an observation is proportional to its mean to the power p. p must
            be greater than 1 and less than or equal to 2. 1 would be Poisson, 2 is gamma.
        link: If a float/int, treated as $\lambda$ in a link function based on
            $\eta = \mu^ \lambda$, meaning 0 gives the log link and 1 gives the
            identity link (i.e. R stats package `power`). Can also be one of "log",
            "identity", "inverse", "sqrt".
    """

    def __init__(
        self,
        p: float | int = 1,
        link: Literal["log", "identity", "inverse", "sqrt"] | int | float = 0,
    ):
        if isinstance(link, int | float):
            link = rstats.power(link)
        self.rfamily = rmgcv.Tweedie(p, link)


class Tw(AbstractFamily):
    r"""Tweedie family with estimated power.

    Restricted to variance function powers between 1 and 2.

    Args:
        link: The link function to use.
        a: The lower bound of the power parameter for optimization.
        b: The upper bound of the power parameter for optimization.
        theta: Related to the Tweedie power parameter by
            $p=(a+b \exp(\theta))/(1+\exp(\theta))$. If this is supplied as a positive
            value then it is taken as the fixed value for p. If it is a negative values
            then its absolute value is taken as the initial value for p.
        theta_fixed: If theta is provided, controls whether to treat theta as fixed
            or estimated. If estimated, then theta is the starting value.
    """

    def __init__(
        self,
        link: Literal["log", "identity", "inverse", "sqrt"] = "log",
        a: float = 1.01,
        b: float = 1.99,
        theta: float | int | None = None,
        *,
        theta_fixed: bool = False,
    ):
        if theta_fixed and theta is None:
            raise ValueError("Theta must be specified if fixed.")

        if theta is not None and not theta_fixed:
            theta = -theta  # mgcv convention.

        self.rfamily = rmgcv.tw(
            theta=ro.NULL if theta is None else theta,
            link=link,
            a=a,
            b=b,
        )


class ZIP(AbstractFamily):
    r"""Zero-inflated Poisson family.

    The probability of a zero count is given by $1-p$, whereas the probability of
    count $y>0$ is given by the truncated Poisson probability function
    $p\mu^y/((\exp(\mu)-1)y!)$. The linear predictor gives $\log \mu$, while
    $\eta = \log(-\log(1-p))$ and $\eta = \theta_1 + \{b+\exp(\theta_2)\} \log \mu$.
    The theta parameters are estimated alongside the smoothing parameters. Increasing
    the b parameter from zero can greatly reduce identifiability problems, particularly
    when there are very few non-zero data.

    The fitted values for this model are the log of the Poisson parameter. Use the
    predict function with type=="response" to get the predicted expected response. Note
    that the theta parameters reported in model summaries are
    $\theta_1 and b + \exp(\theta_2)$.

    !!! warning

        These models should be subject to very careful checking, especially if fitting
        has not converged. It is quite easy to set up models with identifiability
        problems, particularly if the data are not really zero inflated, but simply have
        many zeroes because the mean is very low in some parts of the covariate space.

    Args:
        b: A non-negative constant, specifying the minimum dependence of the zero
            inflation rate on the linear predictor.
        theta: The 2 parameters controlling the slope and intercept of the linear
            transform of the mean controlling the zero inflation rate. If supplied then
            treated as fixed parameters (\theta_1 and \theta_2), otherwise estimated.
    """

    def __init__(
        self,
        b: int | float = 0,
        theta: tuple[int | float, int | float] | None = None,
    ):
        if theta is not None:
            theta = np.asarray(theta)  # type: ignore
        self.rfamily = rmgcv.ziP(theta=ro.NULL if theta is None else theta, b=b)


# TODO support stratification? There is a lot of small details missing in the docs.
# TODO cox.pht
class CoxPH(AbstractFamily):
    """Additive Cox Proportional Hazard Model.

    Cox Proportional Hazards model with Peto's correction for ties, optional
    stratification, and estimation by penalized partial likelihood maximization, for use
    with [`GAM`][pymgcv.gam.GAM]. In the model formula, event time is the response.

    Under stratification the response has two columns: time and a numeric index for
    stratum. The weights vector provides the censoring information (0 for censoring, 1
    for event). CoxPH deals with the case in which each subject has one event/censoring
    time and one row of covariate values.
    """

    def __init__(self):
        self.rfamily = rmgcv.cov_ph()


class GammaLS(AbstractFamily):
    r"""Gamma location-scale model family.

    The log of the mean, $\mu$, and the log of the scale parameter, $\phi$ can depend on
    additive smooth predictors (i.e. using two formulae).

    Args:
        min_log_scale: The minimum value for the log scales parameter.
    """

    def __init__(
        self,
        min_log_scale: float | int = -7,
    ):
        self.family = rmgcv.gammals(b=min_log_scale)


# TODO when e.g. qq plotting finalized check works correctly with GauLSS
class GauLSS(AbstractFamily):
    r"""Gaussian location-scale model family for GAMs.

    Models both the mean $\mu$ and standard deviation $\sigma$ of a Gaussian
    response. The standard deviation uses a "logb" link, i.e.
    $\eta = \log(\sigma - b)$ to avoid singularities near zero.

    Only compatible with [`GAM`][pymgcv.gam.GAM], to which two predictors
    must be specified, for the response variable and the scale respectively.

    - Predictions with `type="response"` returns columns `[mu, 1/sigma]`
    - Predictions with `type="link"` returns columns `[eta_mu, log(sigma - b)]`
    - Plots use the `log(sigma - b)` scale.

    Args:
        link: The link function to use for $\mu$.
        min_std: Minimum standard deviation $b$, for the "logb" link.
    """

    rfamily: ro.ListVector

    def __init__(
        self,
        link: Literal["identity", "inverse", "log", "sqrt"] = "identity",
        min_std: float = 0.01,
    ):
        self.rfamily = rmgcv.gaulss(link=ro.StrVector([link, "logb"]), b=min_std)


class GevLSS(AbstractFamily):
    r"""Generalized extreme value location, scale and shape family.

    Requires three predictors, one for the location, log scale and the shape.

    Uses the p.d.f. $t(y)^{\xi+1} e^{-t(y)} / \sigma$, where:
    $t(x) = [1 + \xi(y-\mu)/\sigma]^{-1/\xi}$ if $\xi \neq 0$
    and $\exp[-(y-\mu)/\sigma]$ otherwise.
    """

    def __init__(self, link: Literal["identity", "log"]):
        self.rfamily = rmgcv.gevlss(link=link)


class GumbLS(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class Multinom(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class MVN(AbstractFamily):
    """Multivariate normal family.

    Args:
        d: The dimension of the distribution.
    """

    rfamily: ro.ListVector

    def __init__(self, d: int):
        self.rfamily = rmgcv.mvn(d=d)


class Shash(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class TwLSS(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()


class ZipLSS(AbstractFamily):
    def __init__(self):
        raise NotImplementedError()



from copy import deepcopy
from typing import Any

import numpy as np
import pandas as pd
import pytest

from pymgcv.gam import BAM, GAM, AbstractGAM
from pymgcv.rpy_utils import to_py
from pymgcv.terms import L
from pymgcv.utils import data_len

from .gam_test_cases import GAMTestCase, get_test_cases, smooth_1d_by_numeric_gam

test_cases = get_test_cases()


@pytest.mark.parametrize("test_case", test_cases.values(), ids=test_cases.keys())
def test_pymgcv_mgcv_equivilance(test_case: GAMTestCase):
    pymgcv_gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    mgcv_gam = test_case.mgcv_gam(test_case.data)
    assert (
        pytest.approx(
            expected=mgcv_gam.rx2["coefficients"],
        )
        == pymgcv_gam.coefficients().to_numpy()
    )


@pytest.mark.parametrize("test_case", test_cases.values(), ids=test_cases.keys())
def test_predict_terms_structure(test_case: GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    all_terms_with_se = gam.partial_effects(test_case.data, compute_se=True)
    expected = test_case.expected_predict_terms_structure

    assert sorted(all_terms_with_se.keys()) == sorted(expected.keys())

    for term_name, fit_and_se in all_terms_with_se.items():
        for fit_or_se in (fit_and_se.fit, fit_and_se.se):
            actual = fit_or_se.columns.values.tolist()
            assert sorted(expected[term_name]) == sorted(actual)


@pytest.mark.parametrize("test_case", test_cases.values(), ids=test_cases.keys())
def test_partial_effects_colsum_matches_predict(test_case: GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    predictions = gam.predict(test_case.data)
    term_predictions = gam.partial_effects(test_case.data, compute_se=True)

    for target, pred in predictions.items():
        term_fit = term_predictions[target].fit
        assert pytest.approx(pred) == term_fit.sum(axis=1)


@pytest.mark.parametrize("test_case", test_cases.values(), ids=test_cases.keys())
def test_partial_effect_against_partial_effects(test_case: GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)

    partial_effects = gam.partial_effects(test_case.data, compute_se=True)

    all_predictors = gam.all_predictors
    for target, terms in all_predictors.items():
        for term in terms:
            try:
                effect = gam.partial_effect(
                    term,
                    target,
                    test_case.data,
                    compute_se=True,
                )
            except NotImplementedError as e:
                if str(e) != "":
                    raise e
                continue

            name = term.label()
            expected_fit = pytest.approx(partial_effects[target].fit[name], abs=1e-6)
            expected_se = pytest.approx(partial_effects[target].se[name], abs=1e-6)

            assert expected_fit == effect.fit
            assert expected_se == effect.se


def test_invalid_type():
    rng = np.random.default_rng(1)
    gam = GAM({"y": L("x")})
    data = pd.DataFrame({"y": rng.normal(size=100), "x": rng.normal(size=100)})
    data["x"] = data["x"].astype(str)
    with pytest.raises(TypeError, match="is of unsupported type"):
        gam = gam.fit(data)

    data = {"x": np.asarray(data["x"]), "y": data["y"]}
    with pytest.raises(TypeError, match="is of unsupported type"):
        gam = gam.fit(data)


@pytest.mark.parametrize("test_case", test_cases.values(), ids=test_cases.keys())
def test_with_se_matches_without(test_case: GAMTestCase):
    gam = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)

    partial_effects_with_se = gam.partial_effects(compute_se=True)
    partial_effects_without = gam.partial_effects(compute_se=False)

    for target in gam.all_predictors.keys():
        assert (
            pytest.approx(partial_effects_with_se[target].fit)
            == partial_effects_without[target]
        )


abstract_method_test_cases = [
    "GAM - smooth_1d_gam",
    "GAM - multivariate_normal_gam",
    "GAM - gaulss_gam",
    "BAM - smooth_1d_random_wiggly_curve_gam",
]


# check it gives something reasonable in both uni/multivariate, and bam case
@pytest.mark.parametrize(
    "test_case",
    [test_cases[k] for k in abstract_method_test_cases],
    ids=abstract_method_test_cases,
)
def test_abstract_methods(test_case: GAMTestCase):
    fit = test_case.gam_model.fit(test_case.data, **test_case.fit_kwargs)
    coef = fit.coefficients()
    cov = fit.covariance()
    assert cov.shape[0] == cov.shape[1]
    assert cov.shape[0] == coef.shape[0]
    assert np.all(coef.index == cov.index)
    assert isinstance(fit.aic(), float)

    residuals = fit.residuals()

    assert residuals.shape[0] == data_len(test_case.data)
    assert fit.fit_state is not None
    mgcv_gam = fit.fit_state.rgam
    resid_from_y_and_fit = fit.residuals_from_y_and_fit(
        y=to_py(mgcv_gam.rx2["y"]),
        fit=to_py(mgcv_gam.rx2["fitted.values"]),
        weights=to_py(mgcv_gam.rx2["prior.weights"]),
    )
    assert np.all(residuals == resid_from_y_and_fit)
    assert isinstance(fit.edf(), pd.Series)
    assert isinstance(fit.penalty_edf(), pd.Series)

    k_check = fit.check_k()
    assert isinstance(k_check, pd.DataFrame)
    expected_columns = ["term", "max_edf", "edf", "k_index", "p_value"]
    assert k_check.columns.to_list() == expected_columns


@pytest.mark.parametrize("gam_type", [GAM, BAM])
@pytest.mark.parametrize("add_nan_to", ["y", "x", "x1"])  # Response, predictor, by
@pytest.mark.parametrize("nan", [np.nan, pd.NA])
def test_errors_on_nans(gam_type: type[AbstractGAM], add_nan_to: str, nan: Any):
    # Erroring on nans seems a sensible default. Strict but avoids
    # possible issues (e.g. length mismatch if unexpectedly dropped).
    test_case: GAMTestCase = smooth_1d_by_numeric_gam(gam_type)

    data = deepcopy(test_case.data)
    gam = test_case.gam_model
    gam.fit(data)
    assert isinstance(data, pd.DataFrame)
    data[add_nan_to] = (
        (data[add_nan_to] * 100).round().astype(pd.Int64Dtype())
    )  # nullable int
    data.at[2, add_nan_to] = nan

    with pytest.raises(ValueError, match="NaN"):
        gam.fit(data)

    if add_nan_to != "y":
        with pytest.raises(ValueError, match="NaN"):
            gam.predict(data)

        with pytest.raises(ValueError, match="NaN"):
            gam.partial_effects(data)

        with pytest.raises(ValueError, match="NaN"):
            gam.partial_effect(gam.predictors["y"][0], data=data)



from dataclasses import dataclass
from typing import Literal

import numpy as np

from pymgcv.families import SupportsCDF
from pymgcv.gam import AbstractGAM
from pymgcv.rlibs import rbase, rstats
from pymgcv.rpy_utils import is_null, to_py, to_rpy


@dataclass
class QQResult:
    """Results required for a qq plot."""

    theoretical: np.ndarray
    residuals: np.ndarray
    interval: tuple[np.ndarray, np.ndarray]  # lower, upper


def qq_simulate(
    gam: AbstractGAM,
    *,
    n_sim: int = 50,
    level: float = 0.9,
    type: Literal["deviance", "response", "pearson"] = "deviance",
) -> QQResult:
    """Generate data for qq-plot via simulation from the family.

    Args:
        gam: The fitted GAM object.
        n_sim: The number of simulations to perform.
        level: The level (i.e. 0.9 means 90% interval).
        type: The type of residuals to use.
    """
    if gam.fit_state is None:
        raise ValueError("GAM must be fitted before simulating quantiles.")
    if n_sim < 2:
        raise ValueError("n must be at least 2.")

    model = gam.fit_state.rgam
    fit = rstats.fitted(model)
    weights = rstats.weights(model, type="prior")
    sigma2 = model.rx2["sig2"]

    if is_null(sigma2):
        sigma2 = rbase.summary(model, re_test=False).rx2["dispersion"]

    fit, weights, sigma2 = to_py(fit), to_py(weights), to_py(sigma2)

    if len(gam.predictors) > 1:
        raise NotImplementedError(
            "Multivariate response families are not supported for qq_simulate.",
        )

    sims = []
    for _ in range(n_sim):
        ysim = gam.family.sample(mu=fit, wt=weights, scale=sigma2)
        res = gam.residuals_from_y_and_fit(
            y=ysim,
            fit=fit,
            weights=weights,
            type=type,
        )
        sim = np.sort(res)
        sims.append(sim)

    sims = np.stack(sims, axis=1)
    n_obs = len(fit)
    theoretical = np.quantile(
        sims,
        q=(np.arange(n_obs) + 0.5) / n_obs,
    )
    alpha = (1 - level) / 2
    interval = np.quantile(sims, q=(alpha, 1 - alpha), axis=1)
    residuals = gam.residuals(type=type)
    residuals = np.sort(residuals)

    return QQResult(
        theoretical=theoretical,
        residuals=residuals,
        interval=(interval[0], interval[1]),
    )


# TODO support normal. For this logp in cdf matters!
def qq_cdf(
    gam: AbstractGAM,
    *,
    level: float = 0.9,
) -> QQResult:
    """Generate a QQ-plot by transforming the data to uniform using the families CDF.

    In this case:

       - The theoretical values are evenly spaced between 0 and 1,
       - The residuals are the result of passing the data through the family's CDF,
         parameterized using the fitted values (and weights and scale as appropriate).

    !!! note

        The Q-Q plot generated by this function likely will not be good at for detecting deviations
        in tail behaviour/due to outliers. For example, a few outliers would be transformed to
        values close to zero or one, which will often **not** be visable in the plot!


        The Q-Q curve when using [`qq_cdf`][pymgcv.qq.qq_cdf] will naturally be a
        different shape to those produced by [`qq_simulate`][pymgcv.qq.qq_simulate]!

    Args:
        gam: The fitted GAM object. The family should support the CDF method,
            which can be checked with `isinstance(family, SupportsCDF)`.
        level: The confidence level for the interval.

    Returns:
        QQResult: The results required for a qq plot.
    """
    if gam.fit_state is None:
        raise ValueError("GAM has not been fit")

    if not isinstance(gam.family, SupportsCDF):
        raise TypeError("Family must support CDF method to use qq_cdf.")

    model = gam.fit_state.rgam

    fit = rstats.fitted(model)
    weights = rstats.weights(model, type="prior")
    sigma2 = model.rx2["sig2"]

    if is_null(sigma2):
        sigma2 = rbase.summary(model, re_test=False).rx2["dispersion"]

    fit, weights, sigma2 = to_py(fit), to_py(weights), to_py(sigma2)

    # Transform the data to uniform (CDF)
    residuals = gam.family.cdf(
        to_py(gam.fit_state.rgam.rx2["y"]),
        mu=fit,
        wt=weights,
        scale=sigma2,
    )
    n_obs = len(residuals)
    theoretical = (np.arange(n_obs) + 0.5) / n_obs
    residuals = np.sort(residuals)
    tmp = np.arange(n_obs)
    alpha = (1 - level) / 2
    lower = rstats.qbeta(alpha, to_rpy(tmp), to_rpy(n_obs - tmp + 1))
    upper = rstats.qbeta(1 - alpha, to_rpy(tmp), to_rpy(n_obs - tmp + 1))

    return QQResult(
        theoretical=theoretical,
        residuals=residuals,
        interval=(lower, upper),
    )
